#include "EXTERN.h"
#include "perl.h"
#include "XSUB.h"

#include <libpkgconf/libpkgconf.h>
#include <stdio.h>

struct my_client_t {
  pkgconf_client_t client;
  FILE *auditf;
  int maxdepth;
  SV *error_handler;
};

typedef struct my_client_t my_client_t;

static bool
#if LIBPKGCONF_VERSION >= 10900
my_error_handler(const char *msg, const pkgconf_client_t *_, void *data)
#else
my_error_handler(const char *msg, const pkgconf_client_t *_, const void *data)
#endif
{
  dSP;

  int count;
  bool value;
  const my_client_t *client = (const my_client_t*)data;
    
  ENTER;
  SAVETMPS;
  
  PUSHMARK(SP);
  EXTEND(SP, 1);
  PUSHs(sv_2mortal(newSVpv(msg, 0)));
  PUTBACK;
  
  count = call_sv(client->error_handler, G_SCALAR);
  
  SPAGAIN;
  
  value = count > 0 && POPi;
  
  PUTBACK;
  FREETMPS;
  LEAVE;

  return value;
}

static bool
my_pkg_iterator(const pkgconf_pkg_t *pkg, void *data)
{
  dSP;
  
  int count;
  bool value;
  SV *callback = (SV*)data;
  
  ENTER;
  SAVETMPS;
  
  PUSHMARK(SP);
  EXTEND(SP,1);
  PUSHs(sv_2mortal(newSViv(PTR2IV(pkg))));
  PUTBACK;
  
  count = call_sv(callback, G_SCALAR);
  
  SPAGAIN;
  
  value = count > 0 && POPi;
  
  PUTBACK;
  FREETMPS;
  LEAVE;
  
  return value;
}

static bool
directory_filter(const pkgconf_client_t *client, const pkgconf_fragment_t *frag, void *data)
{
  if(pkgconf_fragment_has_system_dir(client, frag))
    return false;
  return true;
}

/*
 * Solve cflags/libs recursively using a pkgconf solver for the given package.
 * Type encodes cflags/libs/shared/static queried property.
 * loaded_from_file is true temporarily injecting a loaded-from-file package
 * into a package cache as pkgconf_queue_solve() operates only on the cache
 * and packages in a path.
 * On success returns true and the caller needs to free the filtered_list.
 * Otherwise, returns false and the filterer_list is still untouched or
 * already freed.
 */
static bool
solve_flags(pkgconf_pkg_t *package, my_client_t *client, int type,
      bool loaded_from_file, pkgconf_list_t *filtered_list) {
#if LIBPKGCONF_VERSION >= 10900
  pkgconf_pkg_t dep_graph_root = {
    .id = "",
    .realname = "",
    .flags = PKGCONF_PKG_PROPF_VIRTUAL,
  };
  char query_string[PKGCONF_BUFSIZE];
  pkgconf_list_t query = PKGCONF_LIST_INITIALIZER;
  bool resolved;
#endif
  pkgconf_list_t unfiltered_list = PKGCONF_LIST_INITIALIZER;
  int eflag;
  int flags;
  int old_flags;

#if LIBPKGCONF_VERSION >= 10900
  if (sizeof(query_string) <=
      snprintf(query_string, sizeof(query_string), "%s = %s",
      package->id, package->version))
    false;
  pkgconf_queue_push(&query, query_string);
  if (loaded_from_file)
    loaded_from_file = (NULL == pkgconf_cache_lookup(&client->client, package->id));
  if (loaded_from_file)
    pkgconf_cache_add(&client->client, package);
#endif
  old_flags = flags = pkgconf_client_get_flags(&client->client);
  if(type % 2) {
    flags |= (PKGCONF_PKG_PKGF_MERGE_PRIVATE_FRAGMENTS | PKGCONF_PKG_PKGF_SEARCH_PRIVATE);
  } else {
    flags &= ~(PKGCONF_PKG_PKGF_MERGE_PRIVATE_FRAGMENTS | PKGCONF_PKG_PKGF_SEARCH_PRIVATE);
  }
  pkgconf_client_set_flags(&client->client, flags);
#if LIBPKGCONF_VERSION >= 10900
  resolved = pkgconf_queue_solve(&client->client, &query, &dep_graph_root, client->maxdepth);
  if (loaded_from_file)
    pkgconf_cache_remove(&client->client, package);
  pkgconf_queue_free(&query);
  if (!resolved) {
    pkgconf_solution_free(&client->client, &dep_graph_root);
    false;
  }
#endif
  /*
   * TODO: attribute for max depth (also in the list version below)
   */
  eflag = type > 1
#if LIBPKGCONF_VERSION >= 10900
    /* Depth more than 2 duplicates last cflags word. pkgconf hard-codes 2. */
    ? pkgconf_pkg_cflags(&client->client, &dep_graph_root, &unfiltered_list, 2/*client->maxdepth*/)
    : pkgconf_pkg_libs(&client->client,   &dep_graph_root, &unfiltered_list, client->maxdepth);
#else
    ? pkgconf_pkg_cflags(&client->client, package,         &unfiltered_list, client->maxdepth)
    : pkgconf_pkg_libs(&client->client,   package,         &unfiltered_list, client->maxdepth);
#endif
  pkgconf_client_set_flags(&client->client, old_flags);
  /*
   * TODO: throw an exception (also in the list verson below)
   */
  if(eflag != PKGCONF_PKG_ERRF_OK) {
#if LIBPKGCONF_VERSION >= 10900
    pkgconf_solution_free(&client->client, &dep_graph_root);
#endif
    false;
  }
  pkgconf_fragment_filter(&client->client, filtered_list, &unfiltered_list, directory_filter, NULL);

  pkgconf_fragment_free(&unfiltered_list);
#if LIBPKGCONF_VERSION >= 10900
  pkgconf_solution_free(&client->client, &dep_graph_root);
#endif
  return true;
}

MODULE = PkgConfig::LibPkgConf  PACKAGE = PkgConfig::LibPkgConf::Client


void
_init(object, error_handler, maxdepth)
    SV *object
    SV *error_handler
    int maxdepth
  INIT:
    my_client_t *self;
  CODE:
    Newxz(self, 1, my_client_t);
    self->auditf = NULL;
    self->error_handler = SvREFCNT_inc(error_handler);
    self->maxdepth = maxdepth;
    pkgconf_client_init(&self->client, my_error_handler, self, pkgconf_cross_personality_default());
    pkgconf_client_set_flags(&self->client, PKGCONF_PKG_PKGF_NONE);
    hv_store((HV*)SvRV(object), "ptr", 3, newSViv(PTR2IV(self)), 0);


void
audit_set_log(self, filename, mode)
    my_client_t *self
    const char *filename
    const char *mode
  CODE:
    if(self->auditf != NULL)
    {
      fclose(self->auditf);
      self->auditf = NULL;
    }
    self->auditf = fopen(filename, mode);
    if(self->auditf != NULL)
    {
      pkgconf_audit_set_log(&self->client, self->auditf);
    }
    else
    {
      /* TODO: call error ? */
    }


const char *
sysroot_dir(self, ...)
    my_client_t *self
  CODE:
    if(items > 1)
    {
      pkgconf_client_set_sysroot_dir(&self->client, SvPV_nolen(ST(1)));
    }
    RETVAL = pkgconf_client_get_sysroot_dir(&self->client);
  OUTPUT:
    RETVAL


const char *
buildroot_dir(self, ...)
    my_client_t *self
  CODE:
    if(items > 1)
    {
      pkgconf_client_set_buildroot_dir(&self->client, SvPV_nolen(ST(1)));
    }
    RETVAL = pkgconf_client_get_buildroot_dir(&self->client);
  OUTPUT:
    RETVAL


int
maxdepth(self, ...)
    my_client_t *self
  CODE:
    if(items > 1)
    {
      self->maxdepth = SvIV(ST(1));
    }
    RETVAL = self->maxdepth;
  OUTPUT:
    RETVAL


void
path(self)
    my_client_t *self
  INIT:
    pkgconf_node_t *n;
    pkgconf_path_t *pnode;
    int count = 0;
  CODE:
    PKGCONF_FOREACH_LIST_ENTRY(self->client.dir_list.head, n)
    {
      pnode = n->data;
      ST(count++) = sv_2mortal(newSVpv(pnode->path, 0));
    }
    XSRETURN(count);


void
filter_lib_dirs(self)
    my_client_t *self
  INIT:
    pkgconf_node_t *n;
    pkgconf_path_t *pnode;
    int count = 0;
  CODE:
    PKGCONF_FOREACH_LIST_ENTRY(self->client.filter_libdirs.head, n)
    {
      pnode = n->data;
      ST(count++) = sv_2mortal(newSVpv(pnode->path, 0));
    }
    XSRETURN(count);    


void
filter_include_dirs(self)
    my_client_t *self
  INIT:
    pkgconf_node_t *n;
    pkgconf_path_t *pnode;
    int count = 0;
  CODE:
    PKGCONF_FOREACH_LIST_ENTRY(self->client.filter_includedirs.head, n)
    {
      pnode = n->data;
      ST(count++) = sv_2mortal(newSVpv(pnode->path, 0));
    }
    XSRETURN(count);    


void
DESTROY(self)
    my_client_t *self;
  CODE:
    if(self->auditf != NULL)
    {
      fclose(self->auditf);
      self->auditf = NULL;
    }
    pkgconf_client_deinit(&self->client);
    SvREFCNT_dec(self->error_handler);
    Safefree(self);


IV
_find(self, name)
    my_client_t *self
    const char *name
  CODE:
    RETVAL = PTR2IV(pkgconf_pkg_find(&self->client, name));
  OUTPUT:
    RETVAL


IV
_package_from_file(self, filename)
    my_client_t *self
    const char *filename
  INIT:
    FILE *fp;
    pkgconf_pkg_t *package;
  CODE:
    fp = fopen(filename, "r");
    if(fp != NULL) {
#if LIBPKGCONF_VERSION >= 10900
      package = pkgconf_pkg_new_from_file(&self->client, filename, fp, 0);
#else
      package = pkgconf_pkg_new_from_file(&self->client, filename, fp);
#endif
      RETVAL = PTR2IV(package);
    } else
      RETVAL = 0;
  OUTPUT:
    RETVAL


void
_scan_all(self, sub)
    my_client_t *self
    SV* sub
  CODE:
    pkgconf_scan_all(&self->client, sub, my_pkg_iterator);
        

void
_dir_list_build(self, env_only)
    my_client_t *self
    int env_only
  INIT:
    int old_flags;
  CODE:
    if(env_only)
    {
      old_flags = pkgconf_client_get_flags(&self->client);
      pkgconf_client_set_flags(&self->client, old_flags | PKGCONF_PKG_PKGF_ENV_ONLY);
    }
    pkgconf_client_dir_list_build(&self->client, pkgconf_cross_personality_default());
    if(env_only)
    {
      pkgconf_client_set_flags(&self->client, old_flags);
    }


void
_set_global(self, kv)
    my_client_t *self
    const char *kv
  CODE:
    pkgconf_tuple_define_global(&self->client, kv);


const char *
_get_global(self, key)
    my_client_t *self
    const char *key
  INIT:
    const char *val;
  CODE:
    val = pkgconf_tuple_find_global(&self->client, key);
    if(val == NULL)
      XSRETURN_EMPTY;
    else
      RETVAL = val;
  OUTPUT:
    RETVAL
    


MODULE = PkgConfig::LibPkgConf  PACKAGE = PkgConfig::LibPkgConf::Package


int
refcount(self)
    pkgconf_pkg_t* self
  CODE:
    RETVAL = self->refcount;
  OUTPUT:
    RETVAL


const char *
id(self)
    pkgconf_pkg_t* self
  CODE:
    RETVAL = self->id;
  OUTPUT:
    RETVAL


const char *
filename(self)
    pkgconf_pkg_t* self
  CODE:
    RETVAL = self->filename;
  OUTPUT:
    RETVAL


const char *
realname(self)
    pkgconf_pkg_t* self
  CODE:
    RETVAL = self->realname;
  OUTPUT:
    RETVAL


const char *
version(self)
    pkgconf_pkg_t* self
  CODE:
    RETVAL = self->version;
  OUTPUT:
    RETVAL


const char *
description(self)
    pkgconf_pkg_t* self
  CODE:
    RETVAL = self->description;
  OUTPUT:
    RETVAL


const char *
url(self)
    pkgconf_pkg_t* self
  CODE:
    RETVAL = self->url;
  OUTPUT:
    RETVAL


const char *
pc_filedir(self)
    pkgconf_pkg_t* self
  CODE:
    RETVAL = self->pc_filedir;
  OUTPUT:
    RETVAL


SV *
_get_string(self, client, type, loaded_from_file)
    pkgconf_pkg_t *self
    my_client_t *client
    int type
    bool loaded_from_file
  INIT:
    pkgconf_list_t filtered_list   = PKGCONF_LIST_INITIALIZER;
    char *buffer;
    size_t len;
    bool escape = true;
  CODE:
    if (!solve_flags(self, client, type, loaded_from_file, &filtered_list))
      XSRETURN_EMPTY;
    len = pkgconf_fragment_render_len(&filtered_list, escape, NULL);
    RETVAL = newSV(len == 1 ? len : len-1);
    SvPOK_on(RETVAL);
    buffer = SvPVX(RETVAL);
    pkgconf_fragment_render_buf(&filtered_list, buffer, len, escape, NULL);
    /*
     * Trim trailing null bytes observed in pkgconf-1.9.4. Probably related to
     * 648a2249fcb10bf679bdb587ef2bbddaab3023ad pkgconf commit.
     */
    while (len > 1 && buffer[len-2] == '\0') len--;
    SvCUR_set(RETVAL, len-1);
    /*
     * Append a space if not already there to mimic pkgconf < 1.9 behaviour.
     */
    if (len > 1 && buffer[len-2] != ' ')
      sv_catpvs(RETVAL, " ");
    pkgconf_fragment_free(&filtered_list);
  OUTPUT:
    RETVAL


void
_get_list(self, client, type, loaded_from_file)
    pkgconf_pkg_t *self
    my_client_t *client
    int type
    bool loaded_from_file
  INIT:
    pkgconf_list_t filtered_list   = PKGCONF_LIST_INITIALIZER;
    pkgconf_node_t *node;
    pkgconf_fragment_t *frag;
    int count = 0;
    HV *h;
  CODE:
    if (!solve_flags(self, client, type, loaded_from_file, &filtered_list))
      XSRETURN_EMPTY;
    PKGCONF_FOREACH_LIST_ENTRY(filtered_list.head, node)
    {
      h = newHV();
      frag = node->data;
      if(frag->type)
        hv_store(h, "type", 4, newSVpvf("%c", frag->type), 0);
      else
        hv_store(h, "type",  4, newSVsv(&PL_sv_undef), 0);
      if(frag->data)
        hv_store(h, "data", 4, newSVpv(frag->data, strlen(frag->data)), 0);
      else
        hv_store(h, "data",  4, newSVsv(&PL_sv_undef), 0);
      ST(count++) = newRV_noinc((SV*) h);
    }
    pkgconf_fragment_free(&filtered_list);
    XSRETURN(count);


void
_get_variable(self, key)
    pkgconf_pkg_t *self
    const char *key
  INIT:
    pkgconf_node_t *node;
    pkgconf_tuple_t *tup;
  CODE:
    PKGCONF_FOREACH_LIST_ENTRY(self->vars.head, node)
    {
      tup = node->data;
      if(!strcmp(tup->key, key))
      {
        XSRETURN_PV(tup->value);
      }
    }
    XSRETURN_EMPTY;
      

MODULE = PkgConfig::LibPkgConf  PACKAGE = PkgConfig::LibPkgConf::Util


void
argv_split(src)
    const char *src
  INIT:
    int argc, ret, i;
    char **argv;
  PPCODE:
    ret = pkgconf_argv_split(src, &argc, &argv);
    if(ret == 0)
    {
      for(i=0; i<argc; i++)
      {
        XPUSHs(sv_2mortal(newSVpv(argv[i],0)));
      }
      pkgconf_argv_free(argv);
    }
    else
    {
      croak("error in argv_split");
    }


int
compare_version(a,b)
    const char *a
    const char *b
  INIT:
    int ret;
  CODE:
    ret = pkgconf_compare_version(a,b);
    if(ret < 0)
      RETVAL = -1;
    else if(ret > 0)
      RETVAL = 1;
    else
      RETVAL = 0;
  OUTPUT:
    RETVAL


char *
path_sep()
  CODE:
    RETVAL = PKG_CONFIG_PATH_SEP_S;
  OUTPUT:
    RETVAL


#define STRINGIZE(x) #x
#define STRINGIZE_VALUE_OF(x) STRINGIZE(x)


const char *
version()
  CODE:
    RETVAL = STRINGIZE_VALUE_OF(MY_PKGCONF_VERSION);
  OUTPUT:
    RETVAL


SV *
path_relocate(in)
    const char *in;
  INIT:
    char out[PKGCONF_BUFSIZE];
    bool ok;
  CODE:
    strncpy(out, in, PKGCONF_BUFSIZE-1);
    ok = pkgconf_path_relocate(out, sizeof out);
    RETVAL = newSVpv(ok ? out : in, 0);
  OUTPUT:
    RETVAL


MODULE = PkgConfig::LibPkgConf  PACKAGE = PkgConfig::LibPkgConf::Test


IV
send_error(client, msg)
    my_client_t *client
    const char *msg
  CODE:
    RETVAL = pkgconf_error(&client->client, "%s", msg);
  OUTPUT:
    RETVAL
  

void
send_log(client, msg)
    my_client_t *client
    const char *msg
  CODE:
    pkgconf_audit_log(&client->client, "%s", msg);

